/*
This file is part of web2w.
Copyright 2017 Martin Ruckert

web2w is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

web2w is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with web2w.  If not, see <http://www.gnu.org/licenses/>.
 
Martin Ruckert, Hochschule Muenchen, Lothstrasse 64, 80336 Muenchen
*/
	/*2:*/
	#line 1075 "web2w.w"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <limits.h>
#include <math.h>
#include "web2w.h"
#include "pascal.tab.h"

	/*11:*/
	#line 1273 "web2w.w"

#define MAX_TOKEN_MEM  250000
	/*:11*/	/*30:*/
	#line 1457 "web2w.w"

#define MAX_STRING_MEM 800000
	/*:30*/	/*35:*/
	#line 1541 "web2w.w"

#define MAX_SYMBOL_TABLE  6007 
#define MAX_SYMBOLS       5200 
	/*:35*/	/*52:*/
	#line 1846 "web2w.w"

#define MAX_MODULE_TABLE 1000
	/*:52*/	/*147:*/
	#line 3747 "web2w.w"

#define VARIADIC  -1
	/*:147*/
	#line 1087 "web2w.w"

	/*12:*/
	#line 1277 "web2w.w"

static token token_mem[MAX_TOKEN_MEM]= {{0}};
static int free_tokens= MAX_TOKEN_MEM;
	/*:12*/	/*16:*/
	#line 1305 "web2w.w"

int sequence_no= 0;
	/*:16*/	/*21:*/
	#line 1338 "web2w.w"

static token*first_token;
token*last_token;
	/*:21*/	/*31:*/
	#line 1461 "web2w.w"

static char string_mem[MAX_STRING_MEM];
static int free_strings= MAX_STRING_MEM;
static int current_string= 0;
	/*:31*/	/*36:*/
	#line 1546 "web2w.w"

symbol*symbol_table[MAX_SYMBOL_TABLE]= {NULL};
static symbol symbols[MAX_SYMBOLS]= {{0}};
static int free_symbols= MAX_SYMBOLS;
	/*:36*/	/*44:*/
	#line 1662 "web2w.w"

#define MAX_LOCALS 50
static int locals[MAX_LOCALS];
static symbol*globals[MAX_LOCALS];
static int free_locals= MAX_LOCALS;
	/*:44*/	/*47:*/
	#line 1748 "web2w.w"

#define MAX_WW_STACK 200
static token*ww_stack[MAX_WW_STACK]= {0};
static int ww_sp= 0;
	/*:47*/	/*53:*/
	#line 1850 "web2w.w"

static module module_table[MAX_MODULE_TABLE]= {{0}};
static int free_modules= MAX_MODULE_TABLE;
static module*module_root= NULL;
	/*:53*/	/*60:*/
	#line 1943 "web2w.w"

token*program;
	/*:60*/	/*63:*/
	#line 1986 "web2w.w"

static symbol*omacros= NULL,**omacro_tail= &omacros,
*pmacros= NULL,**pmacro_tail= &pmacros,
*nmacros= NULL,**nmacro_tail= &nmacros;
	/*:63*/	/*74:*/
	#line 2246 "web2w.w"

#define MAX_PPSTACK 40
static struct{
token*next,*end,*link;
int environment;
token*parameter;
}pp_stack[MAX_PPSTACK+1]= {{0}};
static int pp_sp= MAX_PPSTACK;
	/*:74*/	/*105:*/
	#line 2945 "web2w.w"

static int alfanum= 0;
static int comma= 0;
static int columns= 0;
static int spaces= 0;
	/*:105*/	/*127:*/
	#line 3334 "web2w.w"

static int float_constant_no;
	/*:127*/	/*142:*/
	#line 3623 "web2w.w"

static int print_no,printn_no,print_esc_no,printn_esc_no,
print_nl_no,scan_keyword_no;
	/*:142*/	/*153:*/
	#line 3900 "web2w.w"

static symbol*current_macro;
static int current_arg= 0;
	/*:153*/	/*160:*/
	#line 4049 "web2w.w"

int exit_no;
	/*:160*/	/*166:*/
	#line 4167 "web2w.w"

static token*varlist= NULL;
	/*:166*/	/*203:*/
	#line 5059 "web2w.w"

unsigned int param_mask,param_bit;
	/*:203*/	/*212:*/
	#line 5272 "web2w.w"

static int x_over_n_no,xn_over_d_no;
	/*:212*/	/*219:*/
	#line 5359 "web2w.w"

#define MAX_NAME 256
static char basename[MAX_NAME];
	/*:219*/	/*221:*/
	#line 5410 "web2w.w"

int uppercase= 0;
	/*:221*/	/*223:*/
	#line 5437 "web2w.w"

static FILE*w= NULL;
static FILE*pp_out= NULL;
FILE*logfile= NULL;
static char*exports_name= NULL;
	/*:223*/	/*226:*/
	#line 5490 "web2w.w"

int header_section= 0;
	/*:226*/	/*228:*/
	#line 5519 "web2w.w"

int extern_symbols= -1;
	/*:228*/	/*233:*/
	#line 5666 "web2w.w"

debugmode debugflags= dbgnone;
	/*:233*/
	#line 1088 "web2w.w"

	/*68:*/
	#line 2106 "web2w.w"

int is_pascal(token*t)
{switch(t->tag)
{case NL:case INDENT:
case ATPLUS:case ATSLASH:case ATHASH:case ATBAR:
case ATCTL:case METACOMMENT:case CIGNORE:
case WGUBED:case WTINI:case WTATS:
case ATCOMMA:case ATBACKSLASH:case ATSEMICOLON:
return 0;
default:return 1;
}
}

static token*wback(token*t)
{while(true)
{t= t->previous;
if(t->tag==RIGHT)t= t->link;
else if(is_pascal(t))return t;
}
}

token*tail_call(token*eq)
{token*p;
p= wback(eq->link);
if(p->tag==PMACRO)return p;
else return p->next;
}
	/*:68*/	/*106:*/
	#line 2958 "web2w.w"

static void wput(char c)
{alfanum= comma= false;
if(c==' ')spaces++;
else if(c=='\n')fputc('\n',w),columns= spaces= 0;
else
{	/*107:*/
	#line 2979 "web2w.w"

if(spaces>0)
{alfanum= comma= false;
do fputc(' ',w),spaces--,columns++;
while(0<spaces);
}
	/*:107*/
	#line 2964 "web2w.w"

fputc(c,w),columns++;
alfanum= isalnum(c);
comma= c==',';
}
}

static void wputs(char*str)
{while(*str!=0)wput(*str++);
}
	/*:106*/	/*109:*/
	#line 2998 "web2w.w"

static void wprint(char*str)
{if(isalnum(str[0]))
	/*108:*/
	#line 2994 "web2w.w"

if(alfanum||comma)spaces++;
	/*:108*/
	#line 3001 "web2w.w"

wputs(str);
}
	/*:109*/	/*110:*/
	#line 3010 "web2w.w"

static void wprint_int(int i)
{	/*108:*/
	#line 2994 "web2w.w"

if(alfanum||comma)spaces++;
	/*:108*/
	#line 3012 "web2w.w"

	/*107:*/
	#line 2979 "web2w.w"

if(spaces>0)
{alfanum= comma= false;
do fputc(' ',w),spaces--,columns++;
while(0<spaces);
}
	/*:107*/
	#line 3013 "web2w.w"

columns+= fprintf(w,"%d",i);
alfanum= true;
comma= false;
}
	/*:110*/	/*111:*/
	#line 3023 "web2w.w"

static void wprint_str(char*str)
{wput('"'),str++;
while(*str!=0)
{if(str[0]=='\''&&str[1]=='\'')wput('\''),str++;
else if(str[0]=='"'&&str[1]=='"')wputs("\\\""),str++;
else if(str[0]=='\\')wputs("\\\\");
else if(str[0]=='\''&&str[1]==0)wput('"');
else if(str[0]=='"'&&str[1]==0)wput('"');
else if(str[0]=='"')wputs("\\\"");
else wput(str[0]);
str++;
}
}
	/*:111*/	/*112:*/
	#line 3043 "web2w.w"

static void wprint_pre(char*str)
{if(columns!=0)wput('\n');
while(*str!=0)fputc(*str++,w);
fputc('\n',w);
columns= 0;
alfanum= comma= false;
}
	/*:112*/	/*113:*/
	#line 3058 "web2w.w"

static char*token2string(token*t)
{CHECK(t!=NULL,"Unable to convert NULL token to a string");
switch(t->tag)
{default:
if(t->text!=NULL)
return t->text;
else
return"";
	/*41:*/
	#line 1618 "web2w.w"

case ID:case PID:case PCONSTID:
case PARRAYFILETYPEID:case PARRAYFILEID:
case PFUNCID:case PPROCID:
case PDEFVARID:case PDEFPARAMID:case PDEFREFID:case PDEFCONSTID:
case PDEFTYPEID:case PDEFTYPESUBID:case PDEFFUNCID:case CREFID:
case NMACRO:case OMACRO:case PMACRO:
return SYM(t)->name;
	/*:41*/	/*195:*/
	#line 4949 "web2w.w"

case CSEMICOLON:
return";";
	/*:195*/
	#line 3067 "web2w.w"

}
}
	/*:113*/	/*116:*/
	#line 3116 "web2w.w"

static token*wtoken(token*t);
static token*wprint_to(token*t,token*end)
{while(t!=end)
t= wtoken(t);
return t;
}

static token*wskip_to(token*t,token*end)
{while(t!=end)
t= t->next;
return t;
}
	/*:116*/	/*125:*/
	#line 3249 "web2w.w"

static bool following_directive(token*t)
{while(true)
if(WDEBUG<=t->tag&&t->tag<=WGUBED)return true;
else if(t->tag==ATPLUS||t->tag==ATEX||t->tag==ATSEMICOLON||
t->tag==NL||t->tag==INDENT)t= t->next;
else return false;
}
	/*:125*/	/*131:*/
	#line 3383 "web2w.w"

static token*wid(token*t)
{	/*108:*/
	#line 2994 "web2w.w"

if(alfanum||comma)spaces++;
	/*:108*/
	#line 3385 "web2w.w"

if(t->previous->tag==ATEX)
wputs("@!");
wputs(SYM(t)->name);
return t->next;
}
	/*:131*/	/*139:*/
	#line 3581 "web2w.w"

static int is_string(token*arg)
{if(arg->tag==PSTRING||arg->tag==STRING)
return 1;
else if(arg->tag==PID)
{DBG(dbgstring,"Variable %s string (%d) argument in line %d\n",
arg->sym_ptr->name,arg->sym_ptr->is_string,arg->line_no);
if(arg->sym_ptr->is_string>0)
return 1;
}
return 0;
}

static void pstring2n(token*id,token*arg)
{if(arg==NULL||arg->tag==PCOLON||arg->tag==CREFID)return;
if(id->sym_no==print_no&&!is_string(arg))id->sym_no= printn_no;
else if(id->sym_no==print_esc_no&&!is_string(arg))id->sym_no= printn_esc_no;
}
	/*:139*/	/*149:*/
	#line 3793 "web2w.w"

int count_arity(token*open)
{int count= 1;
token*p;
if(open->next->tag==HASH)return 0;
for(p= open->next;p!=open->link;p= p->next)
if(p->tag==PCOMMA)count++;
else if(p->tag==POPEN)p= p->link;
return count;
}
	/*:149*/	/*155:*/
	#line 3923 "web2w.w"

void wprint_args(int from,int to)
{while(true)
{	/*108:*/
	#line 2994 "web2w.w"

if(alfanum||comma)spaces++;
	/*:108*/
	#line 3926 "web2w.w"

wput('A'+from);
from++;
if(from>=to)break;
else wput(',');
}
}
	/*:155*/	/*177:*/
	#line 4402 "web2w.w"

static long int generate_constant(token*t,char sign,long int value)
{if(t->tag==PTYPECHAR||t->tag==PINTEGER)
{if(sign=='-')return value-t->value;else return value+t->value;}
else if(t->tag==NMACRO||t->tag==PCONSTID)
{if(sign!=0)wput(sign);wprint(token2string(t->previous));
return value;
}
if(t->tag==PPLUS)
{if(t->previous!=NULL)value= generate_constant(t->previous,sign,value);
if(sign==0)sign= '+';
return generate_constant(t->next,sign,value);
}
if(t->tag==PMINUS)
{if(t->previous!=NULL)value= generate_constant(t->previous,sign,value);
if(sign==0||sign=='+')sign= '-';else sign= '+';
return generate_constant(t->next,sign,value);
}
else
ERROR("Unexpected tag %s while generating a constant expression in line %d",TAG(t),t->line_no);
}
	/*:177*/	/*185:*/
	#line 4594 "web2w.w"

static int dead_end(token*t,int line_no)
{DBG(dbgbreak,"Searching for dead end in line %d:\n",line_no);
while(true)
{DBG(dbgbreak,"\t%s\n",TAG(t));
if(t->tag==PGOTO||t->tag==PEXIT||t->tag==CPROCRETURN)return true;
else if(t->tag==PCOLON)t= t->next;
else if(t->tag==PBEGIN)t= t->previous;
else if(t->tag==PSEMICOLON||t->tag==CCASE)
{if(t->next->tag==CEMPTY)t= t->previous;else t= t->next;}
else return false;
}
}
	/*:185*/
	#line 1089 "web2w.w"

	/*14:*/
	#line 1286 "web2w.w"

static token*new_token(int tag)
{token*n;
if(free_tokens>0)n= &token_mem[--free_tokens];
else ERROR("token mem overflow");
n->line_no= ww_lineno;
n->sequence_no= sequence_no;
n->tag= tag;
return n;
}
	/*:14*/	/*19:*/
	#line 1321 "web2w.w"

void seq(token*t,token*s)
{CHECK(t->sequence_no==s->sequence_no,"tokens in line %d "
"and %d belong to different code sequences",t->line_no,s->line_no);
}
	/*:19*/	/*24:*/
	#line 1354 "web2w.w"

token*add_token(int tag)
{token*n= new_token(tag);
last_token->next= n;
n->previous= last_token;
last_token= n;
return n;
}
	/*:24*/	/*33:*/
	#line 1476 "web2w.w"

char*new_string(void)
{current_string= MAX_STRING_MEM-free_strings;
return string_mem+current_string;
}

static void add_char(char c)
{if(free_strings>0)
string_mem[MAX_STRING_MEM-free_strings--]= c;
else ERROR("String memory overflow");
}

void add_string(char*str)
{while(*str!=0)add_char(*str++);
}

char*end_string(void)
{char*str= string_mem+current_string;
if(free_strings>0)
string_mem[MAX_STRING_MEM-free_strings--]= 0;
else
ERROR("String memory overflow");
current_string= MAX_STRING_MEM-free_strings;
return str;
}

char*copy_string(char*str)
{new_string();
add_string(str);
return end_string();
}

int string_length(void)
{return(MAX_STRING_MEM-free_strings)-current_string;}

void flush_string(void)
{free_strings= MAX_STRING_MEM-current_string;
if(free_strings>MAX_STRING_MEM)ERROR("String memory underflow");
}

	/*:33*/	/*38:*/
	#line 1561 "web2w.w"

static int symbol_hash(char*name)
{int hash= 0;
while(*name!=0)
hash= hash+(*(name++)^0x9E);
return hash;
}

static symbol*new_symbol(void)
{CHECK(free_symbols>0,"Symbol table overflow");
free_symbols--;
return symbols+free_symbols;
}

int sym_no(char*name)
{int i,c;
i= symbol_hash(name)%MAX_SYMBOL_TABLE;
if(symbol_table[i]!=NULL)
{if(strcmp(symbol_table[i]->name,name)==0)
return i;
if(i==0)c= 1;else c= MAX_SYMBOL_TABLE-i;
while(true)
{i= i-c;
if(i<0)i= i+MAX_SYMBOL_TABLE;
if(symbol_table[i]==NULL)break;
if(strcmp(symbol_table[i]->name,name)==0)
return i;
}
}
symbol_table[i]= new_symbol();
symbol_table[i]->name= copy_string(name);
symbol_table[i]->tag= ID;
return i;
}
	/*:38*/	/*45:*/
	#line 1670 "web2w.w"

void scope_open(void)
{
CHECK(free_locals==MAX_LOCALS,"Opening a new scope without closing the previous one");
}

void scope_close(void)
{int i;
for(i= free_locals;i<MAX_LOCALS;i++)
{globals[i]->use_count= symbol_table[locals[i]]->use_count;
globals[i]->scan_count= symbol_table[locals[i]]->scan_count;
globals[i]->is_label= symbol_table[locals[i]]->is_label;
symbol_table[locals[i]]= globals[i];
}
free_locals= MAX_LOCALS;
}
	/*:45*/	/*46:*/
	#line 1690 "web2w.w"

void localize(token*t)
{int sym_no= t->sym_no;
symbol*l,*g;
l= new_symbol();
g= symbol_table[sym_no];
l->name= g->name;
l->tag= g->tag;
l->eq= g->eq;
l->use_count= g->use_count;
l->scan_count= g->scan_count;
l->is_label= g->is_label;
symbol_table[sym_no]= l;
CHECK(free_locals>0,"Overflow of local symbols in line %d",t->line_no);
free_locals--;
locals[free_locals]= sym_no;
globals[free_locals]= g;
t->sym_ptr= l;
}
	/*:46*/	/*49:*/
	#line 1767 "web2w.w"

void ww_push(token*left)
{CHECK(ww_sp<MAX_WW_STACK,"WW stack overflow");
DBG(dbglink,"Pushing[%d]:",ww_sp);if(left!=NULL)DBG(dbglink,THE_TOKEN(left));
ww_stack[ww_sp++]= left;
}

token*ww_pop(token*right)
{token*left;
CHECK(ww_sp>0,"Mode stack underflow");
left= ww_stack[--ww_sp];
if(left!=NULL)left->link= right;
DBG(dbglink,"Popping[%d]:",ww_sp);if(left!=NULL)DBG(dbglink,THE_TOKEN(left));
return left;
}

int ww_top_is(int tag)
{return ww_sp>0&&ww_stack[ww_sp-1]!=NULL&&ww_stack[ww_sp-1]->tag==tag;
}

	/*:49*/	/*51:*/
	#line 1834 "web2w.w"

static int module_name_cmp(token*n,token*m)
{n= n->next;m= m->next;
if(n->next->tag==ELIPSIS)
return strncmp(n->text,m->text,strlen(n->text));
else
return strcmp(n->text,m->text);
}
	/*:51*/	/*56:*/
	#line 1876 "web2w.w"

module*find_module(token*atless)
{module**m= &module_root;
while(*m!=NULL)
{int d= module_name_cmp(atless,(*m)->atless);
if(d==0)
return*m;
else if(d<0)m= &((*m)->left);
else m= &((*m)->right);
}
CHECK(free_modules>0,"Module table overflow");
*m= module_table+MAX_MODULE_TABLE-free_modules--;
(*m)->atless= atless;
return*m;
}
	/*:56*/	/*57:*/
	#line 1898 "web2w.w"

void add_module(token*atless)
{module*m= find_module(atless);
token*atgreater= m->atgreater;
if(atgreater==NULL)
m->atgreater= atless->link;
else
{while(atgreater->link!=NULL)atgreater= atgreater->link;
atgreater->link= atless->link;
}
}
	/*:57*/	/*64:*/
	#line 1997 "web2w.w"

void def_macro(token*eq,int tag)
{token*id;
if(eq->previous->tag==PARAM)
{id= eq->previous->previous;
tag= PMACRO;
}
else
id= eq->previous;
CHGTAG(id,tag);
CHGID(id,tag);
SYM(id)->eq= eq;
if(tag==NMACRO)
{*nmacro_tail= SYM(id);nmacro_tail= &(SYM(id)->link);}
else if(tag==OMACRO)
{*omacro_tail= SYM(id);omacro_tail= &(SYM(id)->link);}
else if(tag==PMACRO)
{*pmacro_tail= SYM(id);pmacro_tail= &(SYM(id)->link);}
DBG(dbgexpand,"Defining %s: %s\n",tagname(tag),SYM(id)->name);
}
	/*:64*/	/*73:*/
	#line 2226 "web2w.w"

void pp_error(const char*message)
{ERROR("Parse error in line %d: %s",pp_lval->line_no,message);}
	/*:73*/	/*75:*/
	#line 2284 "web2w.w"

token*pp_push(token*link,token*eq,int environment,token*parameter)
{CHECK(pp_sp>0,"Pascal lexer stack overflow");
pp_sp--;
pp_stack[pp_sp].link= link;
pp_stack[pp_sp].next= eq->next;
pp_stack[pp_sp].end= eq->link;
pp_stack[pp_sp].environment= environment;
pp_stack[pp_sp].parameter= parameter;
DBG(dbgexpand,"Push pp_lex[%d]: ",MAX_PPSTACK-pp_sp);DBGTOKS(dbgexpand,eq->next,eq->link);
return pp_stack[pp_sp].next;
}

token*pp_pop(void)
{CHECK(pp_sp<MAX_PPSTACK,"Pascal Lexer stack underflow");
pp_sp++;
DBG(dbgexpand,"Pop pp_lex[%d]: ",MAX_PPSTACK-pp_sp);
DBGTOKS(dbgexpand,pp_stack[pp_sp].next,pp_stack[pp_sp].end);
return pp_stack[pp_sp].next;
}
	/*:75*/	/*76:*/
	#line 2315 "web2w.w"

int pp_lex(void)
{token*t;
int tag;
t= pp_stack[pp_sp].next;
while(true)
{
if(t==pp_stack[pp_sp].end)
{	/*95:*/
	#line 2682 "web2w.w"

token*link= pp_stack[pp_sp].link;
if(link!=NULL)
{token*eq;
eq= link->next;
link= link->link;
pp_pop();
t= pp_push(link,eq,0,NULL);
}
else
t= pp_pop();
	/*:95*/
	#line 2323 "web2w.w"

continue;
}
tag= t->tag;
tag_known:
if(tag>FIRST_PASCAL_TOKEN)
{
pp_stack[pp_sp].next= t->next;
goto found;
}
else
{switch(tag)
{	/*78:*/
	#line 2363 "web2w.w"

case NL:
case INDENT:
if(pp_out!=NULL)fprintf(pp_out,"%s",token2string(t));
case METACOMMENT:
case ATCTL:
case ATEX:
case ATQM:
case ATPLUS:
case ATSLASH:
case ATBACKSLASH:
case ATBAR:
case ATHASH:
case ATCOMMA:
case ATAND:
case ATSEMICOLON:
case ATLEFT:
case ATRIGHT:
t= t->next;
continue;
	/*:78*/	/*79:*/
	#line 2385 "web2w.w"

case MLEFT:
case PLEFT:
t= t->link->next;
continue;
	/*:79*/	/*80:*/
	#line 2394 "web2w.w"

case PEOF:
pp_stack[pp_sp].next= t->next;
goto found;
	/*:80*/	/*81:*/
	#line 2406 "web2w.w"

case ATDOLLAR:
case OCTAL:
case HEX:
case CHAR:
pp_stack[pp_sp].next= t->next;
tag= PINTEGER;
goto found;
	/*:81*/	/*82:*/
	#line 2422 "web2w.w"

case ID:
{symbol*s= SYM(t);
tag= s->tag;
if(tag==ID)
tag= s->tag= PID;
t->sym_ptr= s;
t->tag= tag;
goto tag_known;
}
	/*:82*/	/*87:*/
	#line 2511 "web2w.w"

case WDEBUG:
case WGUBED:
case WINIT:
case WTINI:
case WSTAT:
case WTATS:
t= t->next;
continue;
	/*:87*/	/*90:*/
	#line 2583 "web2w.w"

case NMACRO:
t->sym_ptr= SYM(t);
if(t->sym_ptr->eq->next->tag==STRING)
{token*s= t->sym_ptr->eq->next;
if(s->sym_ptr!=NULL)
{s->sym_ptr= SYM(s);
DBG(dbgstring,"Using numeric macro %s  in line %d\n",
s->sym_ptr->name,t->line_no);
}
}
pp_stack[pp_sp].next= t->next;
goto found;
case STRING:
if(t->sym_no!=0)
{t->sym_ptr= SYM(t);
}
DBG(dbgstring,"Using string %s in line %d\n",
t->text,t->line_no);
pp_stack[pp_sp].next= t->next;
goto found;
	/*:90*/	/*94:*/
	#line 2664 "web2w.w"

case ATLESS:
{token*eq,*atgreater;
atgreater= find_module(t)->atgreater;
CHECK(atgreater!=NULL,"Undefined module @<%s ...@> in line %d",
token2string(t->next),t->line_no);
DBG(dbgexpand,"Expanding module @<%s@> in line %d\n",
token2string(t->next),t->line_no);
eq= atgreater->next;
pp_stack[pp_sp].next= t->link->next;
t= pp_push(atgreater->link,eq,0,NULL);
continue;
}
	/*:94*/	/*96:*/
	#line 2698 "web2w.w"

case OMACRO:
if(SYM(t)->tag==NMACRO)
{tag= NMACRO;
t->sym_ptr= SYM(t);
pp_stack[pp_sp].next= t->next;
goto found;
}
else
{token*eq;
eq= SYM(t)->eq;SYM(t)->use_count++;
pp_stack[pp_sp].next= t->next;
DBG(dbgexpand,"Expanding ordinary macro %s in line %d\n",
token2string(t),t->line_no);
t= pp_push(NULL,eq,0,NULL);
continue;
}
	/*:96*/	/*97:*/
	#line 2751 "web2w.w"

case PMACRO:
{token*open,*eq;
int popen;
DBG(dbgexpand,"Expanding parameter macro %s in line %d\n",
token2string(t),t->line_no);
eq= SYM(t)->eq;SYM(t)->use_count++;
pp_stack[pp_sp].next= t->next;
popen= pp_lex();
CHECK(popen==POPEN,"expected ( after macro with parameter");
open= pp_lval;
pp_stack[pp_sp].next= open->link->next;
	/*148:*/
	#line 3778 "web2w.w"

{token*p;
int count;
	/*150:*/
	#line 3805 "web2w.w"

count= count_arity(open);
if(count>0)
{if(SYM(t)->arity==0)SYM(t)->arity= count;
else if(SYM(t)->arity!=count)SYM(t)->arity= VARIADIC;
}
DBG(dbgmacro,"Counting %s parameters: %d in line %d\n",
SYM(t)->name,count,open->line_no);
	/*:150*/
	#line 3781 "web2w.w"

	/*151:*/
	#line 3821 "web2w.w"

p= open->link->next;
while(p!=NULL)
{if(p->tag==POPEN)
{if(p->link!=NULL&&p->link->tag==PCLOSE)
{int c= count_arity(p);
if(c==0)
{count= 0;break;}
else count= count+c;
p= p->link->next;
}
else
ERROR(") expected matching ( in line %d",p->line_no);
}
else if(!is_pascal(p))p= p->next;
else break;
}
if(count>0)
{
if(SYM(t)->arg_count==0)SYM(t)->arg_count= count;
else if(SYM(t)->arg_count!=count)SYM(t)->arg_count= VARIADIC;
DBG(dbgmacro,"Counting %s arguments: %d in line %d\n",
SYM(t)->name,count,open->link->next->line_no);
if(SYM(t)->arg_count>'Z'-'A'+1)
ERROR("Macro %s with %d>%d arguments found in line %d",
SYM(t)->name,SYM(t)->arg_count,'Z'-'A'+1,open->link->next->line_no);
}
	/*:151*/
	#line 3782 "web2w.w"

}
	/*:148*/
	#line 2763 "web2w.w"

t= pp_push(NULL,eq,pp_sp,open);
continue;
}
	/*:97*/	/*98:*/
	#line 2777 "web2w.w"

case HASH:
{token*parameter= pp_stack[pp_sp].parameter;
int environment= pp_stack[pp_sp].environment;
pp_stack[pp_sp].next= t->next;
t= pp_push(NULL,parameter,pp_stack[environment].environment,
pp_stack[environment].parameter);
continue;
}
	/*:98*/	/*162:*/
	#line 4064 "web2w.w"

case PRETURN:
t->sym_ptr= SYM(t);
pp_stack[pp_sp].next= t->next;
goto found;
	/*:162*/	/*196:*/
	#line 4954 "web2w.w"

case CSEMICOLON:
t= t->next;
continue;
	/*:196*/
	#line 2335 "web2w.w"

default:
ERROR("Unexpected token in pp_lex:"THE_TOKEN(t));
}
}
}
found:
DBG(dbgpascal,"pp_lex: %s->\t",tagname(tag));
DBG(dbgpascal,THE_TOKEN(t));
if(pp_out!=NULL)fprintf(pp_out,"%s ",token2string(t));
t->pp_sp= pp_sp;
pp_lval= t;
return tag;
}
	/*:76*/	/*84:*/
	#line 2445 "web2w.w"

void propagate_use(symbol*s)
{token*t;
if(s->use_count>1)return;
for(t= s->eq->next;t!=s->eq->link;t= t->next)
if(t->tag==NMACRO)
{SYM(t)->use_count++;propagate_use(SYM(t));}
}
	/*:84*/	/*92:*/
	#line 2613 "web2w.w"

long int getval(token*t)
{long int n= 0;
switch(t->tag)
{case ATDOLLAR:n= 0;break;
case PINTEGER:n= strtol(t->text,NULL,10);break;
case OMACRO:
if(SYM(t)->tag!=NMACRO)
ERROR("Unable to get value for OMACRO in line %d",t->line_no);
case NMACRO:
t= SYM(t)->eq;
CHECK(t->tag==EQEQ,"= expected in numeric macro in line %d",t->line_no);
t= t->next;
if(t->tag==POPEN)t= t->next;
if(t->tag==PMINUS){t= t->next;n= -getval(t);}
else if(t->tag==PPLUS){t= t->next;n= getval(t);}
else n= getval(t);
while(true)
{if(t->next->tag==PPLUS)
{t= t->next->next;n= n+getval(t);}
else if(t->next->tag==PMINUS)
{t= t->next->next;n= n-getval(t);}
else break;
}
break;
case OCTAL:n= strtol(t->text+2,NULL,8);break;
case HEX:n= strtol(t->text+2,NULL,16);break;
case CHAR:n= (int)(unsigned char)t->text[1];break;
case PCONSTID:n= SYM(t)->value;break;
default:ERROR("Unable to get value for tag %s in line %d",TAG(t),t->line_no);
}
return n;
}
	/*:92*/	/*102:*/
	#line 2825 "web2w.w"

token*join(int tag,token*left,token*right,long int value)
{token*n= new_token(tag);
n->previous= left;
n->next= right;
n->value= value;
DBG(dbgjoin,"tree: ");DBGTREE(dbgjoin,n);
return n;
}
	/*:102*/	/*104:*/
	#line 2914 "web2w.w"

int predefine(char*name,int tag,int value)
{int n= sym_no(name);
symbol*s= symbol_table[n];
s->tag= tag;
s->value= value;
return n;
}
	/*:104*/	/*115:*/
	#line 3099 "web2w.w"

static token*wtoken(token*t)
{static int level= 0;
level++;
DBG(dbgcweb,"wtoken[%d] %s (%s) line %d\n",level,TAG(t),token2string(t),t->line_no);
switch(t->tag)
{	/*88:*/
	#line 2525 "web2w.w"

case WDEBUG:
if(t->previous->tag==BAR)
wputs(t->text);
else
{wprint_pre("#ifdef @!DEBUG");
}
t= t->next;break;
case WINIT:
if(t->previous->tag==BAR)
wputs(t->text);
else
{wprint_pre("#ifdef @!INIT");
}
t= t->next;break;
case WSTAT:
if(t->previous->tag==BAR)
wputs(t->text);
else
{wprint_pre("#ifdef @!STAT");
}
t= t->next;break;
case WGUBED:
case WTINI:
case WTATS:
if(t->previous->tag==BAR)
wputs(t->text);
else
{wprint_pre("#endif");
}
t= t->next;
if(t->tag==ATPLUS||t->tag==ATSLASH)t= t->next;
if(t->tag==NL)t= t->next;
break;
	/*:88*/	/*89:*/
	#line 2562 "web2w.w"

case ATPLUS:
t= t->next;
if(!following_directive(t))
wputs("@+");
else
DBG(dbgcweb,"Eliminating @+ in line %d\n",t->line_no);
break;
	/*:89*/	/*117:*/
	#line 3136 "web2w.w"

case CIGNORE:case CTLOCAL:case PLABEL:case PVAR:
case PPACKED:case POF:case ATQM:case ATBACKSLASH:
t= t->next;break;
	/*:117*/	/*119:*/
	#line 3150 "web2w.w"

case PFORWARD:
if(t->next->tag==PSEMICOLON)wput(';'),t= t->next->next;
else wprint("forward"),t= t->next;
if(t->tag==ATCTL)t= t->next;
break;
	/*:119*/	/*120:*/
	#line 3164 "web2w.w"

case METACOMMENT:
{char*c;
wputs(" /*");
for(c= t->text+2;c[0]!='@'||c[1]!='}';c++)wput(*c);
wputs("*/");
t= t->next;
}
break;
case ATLEFT:
wprint_pre("#if 0");
t= t->next;break;
case ATRIGHT:
wprint_pre("#endif");
t= t->next;break;
	/*:120*/	/*121:*/
	#line 3186 "web2w.w"

case PLEFT:case MLEFT:wputs(" /*");t= t->next;break;
case RIGHT:wputs("*/");t= t->next;break;
case PSEMICOLON:wputs(";");t= t->next;break;
case PCOMMA:wputs(",");t= t->next;break;
case PMOD:wput('%');t= t->next;break;
case PDIV:wput('/');t= t->next;break;
case PNIL:wprint("NULL");t= t->next;break;
case POR:wputs("||");t= t->next;break;
case PAND:wputs("&&");t= t->next;break;
case PNOT:wputs("!");t= t->next;break;
case PIF:wprint("if (");t= t->next;break;
case PTHEN:wputs(") ");t= t->next;break;
case PASSIGN:wput('=');t= t->next;break;
case PNOTEQ:wputs("!=");t= t->next;break;
case PEQ:wputs("==");t= t->next;break;
case EQEQ:wput(' ');t= t->next;break;
case OCTAL:wprint("0");wputs(t->text+2);t= t->next;break;
case HEX:wprint("0x");wputs(t->text+2);t= t->next;break;
case PTYPEINT:wprint("int");t= t->next;break;
case PTYPEREAL:wprint("double");t= t->next;break;
case PTYPEBOOL:wprint("bool");t= t->next;break;
case PTYPECHAR:wprint("unsigned char");t= t->next;break;
	/*:121*/	/*122:*/
	#line 3213 "web2w.w"

case PPROGRAM:
if(t->link!=NULL)t= t->link->next;
else wputs(t->text),t= t->next;
break;
	/*:122*/	/*124:*/
	#line 3243 "web2w.w"

case PBEGIN:wput('{'),t= t->next;
if(!following_directive(t))wputs("@+");
break;
	/*:124*/	/*126:*/
	#line 3264 "web2w.w"

case PDOTDOT:wprint("dotdot");t= t->next;break;
	/*:126*/	/*129:*/
	#line 3342 "web2w.w"

case PSLASH:
wput('/');
if(t->next->tag!=PMACRO||t->next->sym_no!=float_constant_no)
{wprint("(double)");
DBG(dbgslash,"Inserting (double) after / in line %d\n",t->line_no);
}
t= t->next;break;
	/*:129*/	/*130:*/
	#line 3365 "web2w.w"

case ATEX:
t= t->next;
if(t->tag!=ID&&t->tag!=PID&&t->tag!=PFUNCID&&
t->tag!=PDEFVARID&&t->tag!=PDEFPARAMID&&t->tag!=PDEFTYPEID&&
t->tag!=OMACRO&&t->tag!=PMACRO&&t->tag!=NMACRO&&
t->tag!=CINTDEF&&t->tag!=CSTRDEF&&t->tag!=PDIV&&
t->tag!=WDEBUG&&t->tag!=WINIT&&t->tag!=WSTAT)
{wputs("@!");
DBG(dbgcweb,"Tag after @! is %s in line %d\n",tagname(t->tag),t->line_no);
}
break;
	/*:130*/	/*132:*/
	#line 3393 "web2w.w"

case ID:
case PID:
case NMACRO:
case CINTDEF:
case PFUNCID:
t= wid(t);break;
	/*:132*/	/*134:*/
	#line 3450 "web2w.w"

case ATLESS:
wputs("@<");
t= t->next;
CHECK(t->tag==TEXT,"Module name expected instead of %s in line %d",
token2string(t),t->line_no);
{char*str= t->text;
do
if(str[0]=='@'&&str[1]==',')
str= str+2;
else if(columns>80&&isspace(*str))wput('\n'),str++;
else wput(*str++);
while(*str!=0);
}
t= t->next;
if(t->tag==ELIPSIS)wputs("..."),t= t->next;
CHECK(t->tag==ATGREATER,"@> expected instead of %s in line %d",
token2string(t),t->line_no);
wputs("@>");
t= t->next;
if(t->tag==ATSLASH)t->tag= ATSEMICOLON;
else if(t->tag==PELSE||
(t->tag==NL))wputs("@;");
break;
	/*:134*/	/*135:*/
	#line 3488 "web2w.w"

case PSTRING:
wprint_str(t->text);
t= t->next;
break;
case CHAR:
case PCHAR:
wput('\'');
switch(t->text[1])
{case'\'':wputs("\\'");break;
case'\\':wputs("\\\\");break;
case'@':wputs("@@");break;
default:wput(t->text[1]);break;
}
wput('\'');
t= t->next;
break;
	/*:135*/	/*136:*/
	#line 3517 "web2w.w"

case STRING:wprint_str(t->text);t= t->next;break;
case ATDOLLAR:wputs("0");t= t->next;break;
	/*:136*/	/*144:*/
	#line 3659 "web2w.w"

case ATD:
{token*eq= t->next->next;
if(eq->tag!=EQEQ)eq= eq->next;
DBG(dbgcweb,"Macro definition in line %d\n",t->line_no);
if(SYM(t->next)->is_label)
{DBG(dbgid,"Eliminating label definition %s in line %d\n",
SYM(t->next)->name,t->line_no);
t= wskip_to(t,eq->link);
}
else if(SYM(t->next)->use_count==0&&SYM(t->next)->scan_count==0)
{DBG(dbgid,"Eliminating unused macro %s in line %d\n",
SYM(t->next)->name,t->line_no);
t= wskip_to(t,eq->link);
}
else
{wputs("@d "),t= t->next;
	/*222:*/
	#line 5414 "web2w.w"

if(uppercase)
{char*s= SYM(t)->name;
while(*s!=0)
{if('a'<=*s&&*s<='z')*s= *s-'a'+'A';
s++;
}
}
	/*:222*/
	#line 3676 "web2w.w"

wprint(SYM(t)->name);
if(t->tag==NMACRO)
	/*145:*/
	#line 3712 "web2w.w"

{wput(' ');
t= eq->next;
}
	/*:145*/
	#line 3679 "web2w.w"

else if(t->tag==OMACRO)
	/*146:*/
	#line 3722 "web2w.w"

{token*p= tail_call(eq);
if(p->tag==PMACRO&&SYM(p)->arg_count>0&&SYM(p)->arity>0)
{int arg_count= SYM(p)->arg_count;
SYM(t)->arg_count= arg_count;
SYM(t)->arity= SYM(p)->arity;
t->tag= SYM(t)->tag= PMACRO;
wput('('),wprint_args(0,arg_count),wputs(") ");
t= wprint_to(eq->next,p->next);
wput('('),wprint_args(0,arg_count),wputs(")");
}
else
{wput(' ');t= eq->next;}
}
	/*:146*/
	#line 3681 "web2w.w"

else if(t->tag==PMACRO)
	/*152:*/
	#line 3853 "web2w.w"

{int arity,arg_count;
current_macro= SYM(t);current_arg= 0;
arity= current_macro->arity,arg_count= current_macro->arg_count;
if(arity==0)
{DBG(dbgmacro,"Defining parametrized macro %s without arguments in line %d\n",
SYM(t)->name,t->line_no);
wputs("(X) ");t= eq->next;
}
else if(arity==VARIADIC)
{DBG(dbgmacro,"Defining variadic macro %s in line %d\n",
SYM(t)->name,t->line_no);
wput('('),wputs("..."),wputs(") ");t= eq->next;
}
else if(arity>arg_count)
ERROR("Macro %s with arity %d > argument count %d in line %d",
SYM(t)->name,arity,arg_count,t->line_no);
else if(arity==arg_count)
{wput('('),wprint_args(0,arg_count),wputs(") ");
t= wprint_to(eq->next,eq->link);
}
else
{token*tail;
DBG(dbgmacro,"Defining macro %s with tail call in line %d\n",
SYM(t)->name,t->line_no);
wput('('),wprint_args(0,arg_count),wputs(") ");
eq= current_macro->eq;
tail= current_macro->type;
t= tail->next;
while(current_arg<arg_count)
{wprint_to(eq->next,tail);
current_arg= current_arg+current_macro->arity;
if(tail->tag!=PMACRO)
break;
current_macro= SYM(tail);
eq= current_macro->eq;
tail= tail_call(eq);
}
}
}
	/*:152*/
	#line 3683 "web2w.w"

else
ERROR("Macro name expected in line %d",t->line_no);
}
DBG(dbgcweb,"End Macro definition in line %d\n",t->line_no);
break;
}
case ATF:
{token*eq= t->next->next;
DBG(dbgcweb,"Format definition in line %d\n",t->line_no);
if(SYM(t->next)->scan_count==0)
{DBG(dbgid,"eliminating unused format definition %s in line %d\n",
SYM(t->next)->name,t->line_no);
t= wskip_to(t,eq->link);
}
else
{wputs("@f "),t= t->next;
wprint(SYM(t)->name);
t= wprint_to(eq->next,eq->link);
}
break;
}
	/*:144*/	/*154:*/
	#line 3907 "web2w.w"

case HASH:
if(current_macro->arity==0)
wprint("X");
else if(current_macro->arity==VARIADIC)
wprint("__VA_ARGS__");
else wprint_args(current_arg,current_arg+current_macro->arity);
t= t->next;break;
	/*:154*/	/*156:*/
	#line 3946 "web2w.w"

case PMACRO:
case OMACRO:
{int count= SYM(t)->arg_count;
token*macro= t;
t= wid(t);
if(count>0&&t->tag==POPEN)
{wput('(');
while(count>0&&t->tag==POPEN)
{token*p= t->next;
count--;
while(p!=t->link)
{if(p->tag==PCOMMA)count--;
else if(p->tag==HASH&&current_macro->arity>0)
count= count-(current_macro->arity-1);
if(p->tag==POPEN)p= wprint_to(p,p->link);
else p= wtoken(p);
}
t= t->link;
if(t->tag!=PCLOSE)
ERROR(") expected in macro call in line %d",t->line_no);
t= t->next;
if(count>0)
{wput(',');
while(!is_pascal(t))t= wtoken(t);
if(t->tag!=POPEN)
ERROR("( expected to continue macro %s(%d) parameters in line %d",
SYM(macro)->name,SYM(macro)->arg_count,t->line_no);
}
}
wput(')');
}
}
break;
	/*:156*/	/*163:*/
	#line 4081 "web2w.w"

case CLABEL:
if(t->sym_ptr->goto_count<=0)
{t= t->next;
if(t->tag==PPLUS)
t= t->next->next;
if(t->tag==PCOLON)
{t= t->next;
if(t->tag==CSEMICOLON)
t= t->next;
}
}
else
{wprint(SYM(t)->name);t= t->next;
if(t->tag==PPLUS)
{t= t->next;wputs(t->text);t= t->next;}
}
break;
case CLABELN:
wprint("label");wputs(t->text);t= t->next;
break;
case PEXIT:
wprint("exit(0)");t= t->next;
break;
case PRETURN:
wprint("goto end");t= t->next;
break;
	/*:163*/	/*164:*/
	#line 4114 "web2w.w"

case PCONST:
wprint("enum {");t= wtoken(t->next);wprint("};");
break;
	/*:164*/	/*165:*/
	#line 4130 "web2w.w"

case CSTRDEF:
t= t->link->next;
while(t->tag==NL)
{t= t->next;
if(t->tag==INDENT)t= t->next;
if(t->tag==PLEFT)t= t->link->next;
if(t->tag==ATCTL)t= t->next;
}
wputs("@!empty_string=256 ""/*the empty string follows after 256 characters*/\n\n");
break;
	/*:165*/	/*167:*/
	#line 4171 "web2w.w"

case PDEFVARID:
DBG(dbgcweb,"Converting variable list in line %d\n",t->line_no);
varlist= t;
{	/*168:*/
	#line 4192 "web2w.w"

symbol*s;
int is_extern= 0,string_type= 0,is_int= 0,is_global= 0;
do{s= t->sym_ptr;
is_global|= s->is_global;
is_extern|= s->is_extern;
is_int|= s->is_int;
if(s->is_string>0)
{DBG(dbgstring,"Variable %s string (%d) defined in line %d\n",
s->name,s->is_string,t->line_no);
string_type= 1;
}
t= t->link;
}while(t->tag==PID);
	/*:168*/
	#line 4175 "web2w.w"

if(!is_extern&&is_global)wprint("static");
	/*169:*/
	#line 4213 "web2w.w"

if(string_type)wprint("char *");
else if(is_int)wprint("int");
else wprint_to(t,t->link);
	/*:169*/
	#line 4177 "web2w.w"

DBG(dbgcweb,"Finished variable type in line %d\n",t->line_no);
while(varlist->tag==PDEFVARID||varlist->tag==PID)
{wid(varlist);
varlist= wprint_to(varlist->next,varlist->link);
}
t= t->link;
}
DBG(dbgcweb,"Finishing variable list in line %d\n",t->line_no);
break;
	/*:167*/	/*170:*/
	#line 4231 "web2w.w"

case PDEFTYPEID:
{token*type_name= t;
token*type= type_name->link;
DBG(dbgcweb,"Defining type %s in line %d\n",token2string(t),t->line_no);
wprint("typedef ");
t= wprint_to(type,type->link);
wprint(token2string(type_name));
break;
}
	/*:170*/	/*171:*/
	#line 4255 "web2w.w"

case CTSUBRANGE:
{long int lo= t->up->previous->value;
long int hi= t->up->next->value;
DBG(dbgcweb,"Defining subrange type %ld..%ld in line %d\n",lo,hi,t->line_no);
if(INT8_MIN<=lo&&hi<=INT8_MAX)wprint("int8_t");
else if(0<=lo&&hi<=UINT8_MAX)wprint("uint8_t");
else if(INT16_MIN<=lo&&hi<=INT16_MAX)wprint("int16_t");
else if(0<=lo&&hi<=UINT16_MAX)wprint("uint16_t");
else if(INT32_MIN<=lo&&hi<=INT32_MAX)wprint("int32_t");
else if(0<=lo&&hi<=UINT32_MAX)wprint("uint32_t");
else ERROR("unable to convert subrange type %ld..%ld in line %d\n",lo,hi,t->line_no);
t= t->link;
break;
}
case CTINT:wprint("int");t= t->link;
break;
	/*:171*/	/*173:*/
	#line 4284 "web2w.w"

case PRECORD:
{DBG(dbgcweb,"Converting record type in line %d\n",t->line_no);
wprint("struct { ");
t= wprint_to(t->next,t->link);
DBG(dbgcweb,"Finished record type in line %d\n",t->line_no);
wput('}');if(t->tag!=NL)wput(' ');
break;
}
case CUNION:
{DBG(dbgcweb,"Converting union type in line %d\n",t->line_no);
wprint("union { ");
t= wprint_to(t->next,t->link);
wprint("};");
DBG(dbgcweb,"Finished union type in line %d\n",t->line_no);
break;
}
	/*:173*/	/*174:*/
	#line 4328 "web2w.w"

case PARRAY:
if(t->up==NULL)
wputs(t->text),t= t->next;
else
{token*from= t->link;
token*index= from->link;
token*to= index->link;
token*element_type= to->link;
token*subrange= t->up;
long int lo,hi;
int zero_based;

if(subrange->tag==PID)subrange= subrange->sym_ptr->type;
lo= subrange->previous->value;
hi= subrange->next->value;

zero_based= (subrange->previous->tag==PINTEGER&&lo==0)||
subrange->previous->tag==PTYPECHAR;
DBG(dbgarray,"Converting array[%ld..%ld] type in line %d\n",lo,hi,t->line_no);
t= wprint_to(element_type,element_type->link);
while(true)
{CHECK(varlist!=NULL,"Nonempty variable list expected in line %d",
varlist->line_no);
DBG(dbgarray,"Generating array variable %s in line %d\n",
varlist->sym_ptr->name,varlist->line_no);
wid(varlist);varlist->sym_ptr->is_zero_based= zero_based;
if(!zero_based)wput('0');
wput('[');	/*175:*/
	#line 4374 "web2w.w"

{int hi,size;
hi= generate_constant(subrange->next,0,0);
size= generate_constant(subrange->previous,'-',hi);
size= size+1;
if(size<0)wput('-'),wprint_int(-size);
else if(size>0)
{if(subrange->previous->tag!=PTYPECHAR&&
(subrange->previous->tag!=PINTEGER||subrange->next->tag!=PINTEGER))wput('+');
wprint_int(size);
}
}
	/*:175*/
	#line 4356 "web2w.w"
wput(']');
if(!zero_based)
{DBG(dbgarray,"Generating array pointer %s[%s=%ld.. ] in line %d\n",
varlist->sym_ptr->name,token2string(from->next),lo,varlist->line_no);
wputs(", *const ");
wid(varlist);
wputs(" = ");
wid(varlist),wput('0');
	/*176:*/
	#line 4388 "web2w.w"

{int lo= generate_constant(subrange->previous,'-',0);
if(lo<0)wput('-'),wprint_int(-lo);
else if(lo>0)wput('+'),wprint_int(lo);
}
	/*:176*/
	#line 4364 "web2w.w"
;
}
varlist= varlist->link;
if(varlist->tag==PDEFVARID||varlist->tag==PID)wput(',');else break;
}
DBG(dbgarray,"Finished array type in line %d\n",t->line_no);
}
break;
	/*:174*/	/*178:*/
	#line 4435 "web2w.w"

case PFILE:
{DBG(dbgcweb,"Converting file type in line %d\n",t->line_no);
wprint("struct {@+FILE *f;@+");
t= wprint_to(t->next,t->link);
wprint("@,d;@+} ");
DBG(dbgcweb,"Finished file type in line %d\n",t->line_no);
break;
}
	/*:178*/	/*179:*/
	#line 4461 "web2w.w"

case PUP:wputs(".d");t= t->next;break;
	/*:179*/	/*180:*/
	#line 4479 "web2w.w"

case PWHILE:
wprint("while ");
if(t->link!=NULL){
wput('(');t= wprint_to(t->next,t->link);wputs(") ");
}
t= t->next;break;
	/*:180*/	/*181:*/
	#line 4497 "web2w.w"

case PCASE:
if(t->link==NULL)
{wprint(t->text);t= t->next;}
else
{wprint("switch (");
t= wprint_to(t->next,t->link);
wputs(") {");
}
break;
	/*:181*/	/*184:*/
	#line 4578 "web2w.w"

case CBREAK:
if(t->previous->tag!=PSEMICOLON&&t->previous->tag!=CSEMICOLON
&&t->previous->tag!=PEND)wputs("@;");
if(!dead_end(t->up,t->line_no))wprint("@+break;");
t= t->next;
break;
	/*:184*/	/*186:*/
	#line 4612 "web2w.w"

case POTHERS:
wprint("default:");
t= t->next;
break;
	/*:186*/	/*187:*/
	#line 4636 "web2w.w"

case PREPEAT:wprint("@/do@+{");t= t->next;break;
case PUNTIL:
{int sequence_no,line_no;
token*end;
wputs("}@+ while (!(");
sequence_no= t->sequence_no;
line_no= t->line_no;
end= t->next;
while(true)
{if(end->tag==PSEMICOLON||end->tag==CSEMICOLON||end->tag==PELSE)
break;
else if(end->tag==ATSPACE||end->tag==ATSTAR)
{end= wback(end)->next;
break;
}
end= end->next;
}
CHECK(sequence_no==end->sequence_no,"until: end of expression not found in line %d",line_no);
t= wprint_to(t->next,end);
wputs("))");
break;
}
	/*:187*/	/*188:*/
	#line 4669 "web2w.w"

case PFOR:
{token*id= t->next;
token*to= t->link;
if(to==NULL){wprint("for");t= t->next;break;}
wprint("for (");
wprint_to(id,to);
wputs("; ");
wid(id);
if(to->tag==PTO)wputs("<=");
else if(to->tag==PDOWNTO)wputs(">=");
else ERROR("to or downto expected in line %d",to->line_no);
wprint_to(to->next,to->link);
wputs("; ");
wid(id);
if(to->tag==PTO)wputs("++");
else wputs("--");
wputs(") ");
t= to->link->next;
break;
}
	/*:188*/	/*191:*/
	#line 4832 "web2w.w"

case PEND:wputs("} "),t= t->next;
if(t->tag==PSEMICOLON)t= t->next;
break;
	/*:191*/	/*197:*/
	#line 4967 "web2w.w"

case PPROCEDURE:
DBG(dbgcweb,"Converting procedure heading in line %d\n",t->line_no);
if(SYM(t)==NULL||SYM(t)->is_extern==0)wprint("static");
	/*198:*/
	#line 4975 "web2w.w"

{wprint("void");
t= wprint_to(t->next,t->link);
if(t->tag!=POPEN)wputs("(void)");
else t= wprint_to(t,t->link->next);
}
	/*:198*/
	#line 4971 "web2w.w"

break;
	/*:197*/	/*199:*/
	#line 4993 "web2w.w"

case PDEFPARAMID:
case PDEFREFID:
{token*varlist= t,*type= t->link;
DBG(dbgcweb,"Converting parameter list in line %d\n",t->line_no);
while(type->tag==PDEFPARAMID||type->tag==PDEFREFID)type= type->link;
while(true)
{wprint_to(type,type->link);
if(varlist->tag==PDEFREFID)wputs(" *");
wid(varlist);
varlist= varlist->link;
if(varlist!=type)wput(',');
else break;
}
t= type->link;
DBG(dbgcweb,"Finishing parameter list in line %d\n",t->line_no);
break;
}
	/*:199*/	/*200:*/
	#line 5018 "web2w.w"

case CREFID:
wputs("(*"),t= wid(t),wput(')');
break;
	/*:200*/	/*201:*/
	#line 5028 "web2w.w"

case CPROCRETURN:
if(t->sym_ptr->goto_count>0)wprint("goto end");
else wprint("return");
t= t->next;
break;
	/*:201*/	/*204:*/
	#line 5073 "web2w.w"

case PCALLID:
DBG(dbgcweb,"Converting call to %s in line %d\n",SYM(t)->name,t->line_no);
pstring2n(t,t->up);
wid(t);
if(SYM(t)->value==1)wputs("()");
t= t->next;break;
	/*:204*/	/*205:*/
	#line 5088 "web2w.w"

case POPEN:
wput('(');
if(t->up==NULL||SYM(t->up)->value==0)
t= t->next;
else
{int param_mask= SYM(t->up)->value;
token*close= t->link;
t= t->next;
if(param_mask<0)wput('&');
param_mask= param_mask<<1;
while(t!=close)
{if(t->tag==PCOMMA)
{wputs(", ");t= t->next;
if(param_mask<0)wput('&');
param_mask= param_mask<<1;
}
else if(t->tag==POPEN)
t= wprint_to(t,t->link);
else
t= wtoken(t);
}
}
break;
	/*:205*/	/*206:*/
	#line 5124 "web2w.w"

case PFUNCTION:
DBG(dbgcweb,"Converting function heading in line %d\n",t->line_no);
if(SYM(t)==NULL||SYM(t)->is_extern==0)wprint("static");
	/*207:*/
	#line 5132 "web2w.w"

{token*param= t->link;
token*type;
if(param->tag==POPEN)type= param->link->link;
else type= param;
wprint_to(type,type->link);
wprint_to(t->next,t->link);
if(param->tag!=POPEN)wputs("(void)");
else wprint_to(param,param->link->next);
t= type->link;
}
	/*:207*/
	#line 5128 "web2w.w"

break;
	/*:206*/	/*211:*/
	#line 5256 "web2w.w"

case CRETURN:
DBG(dbgcweb,"Converted function return %s in line %d\n",SYM(t)->name,t->line_no);
wprint("return");
t= t->next;
break;
	/*:211*/	/*216:*/
	#line 5306 "web2w.w"

case PFBEGIN:
DBG(dbgcweb,"Adding scaled %s; in line %d\n",SYM(t)->name,t->line_no);
wprint("scaled");t= wid(t);wputs(";\n");
break;
case PFEND:
DBG(dbgcweb,"Adding return %s; in line %d\n",SYM(t)->name,t->line_no);
wprint("return");t= wid(t);wputs(";}");
break;
	/*:216*/	/*217:*/
	#line 5326 "web2w.w"

case CMAIN:if(SYM_PTR("main")->is_extern)SYM_PTR("main")->type= t;
else wprint("static");
wprint("int main(void) {");t= t->next;break;
case CMAINEND:wprint("return 0; }");t= t->next;break;
	/*:217*/
	#line 3105 "web2w.w"

default:wprint(token2string(t));t= t->next;break;
}
level--;
return t;
}
	/*:115*/	/*138:*/
	#line 3547 "web2w.w"

void pvar_string(token*id,token*str)
{if(id->tag==PID)
{if(str->tag==STRING||(str->tag==PID&&str->sym_ptr->is_string>0))
id->sym_ptr->is_string++;
else if(str->tag==PINTEGER&&getval(str)==0)
;
else
id->sym_ptr->is_string--;
DBG(dbgstring,"Variable %s string (%d) marked in line %d\n",
id->sym_ptr->name,id->sym_ptr->is_string,id->line_no);
}
}
	/*:138*/	/*141:*/
	#line 3612 "web2w.w"

void pchar2string(token*id,token*arg)
{if(arg->tag==CHAR&&
(id->sym_no==print_no||id->sym_no==print_esc_no||
id->sym_no==print_nl_no||id->sym_no==scan_keyword_no))
arg->tag= PSTRING;
id->up= arg;
}
	/*:141*/	/*158:*/
	#line 4014 "web2w.w"

void clabel(token*t,int reference)
{
if(t->tag==NMACRO)
{t->tag= CLABEL;
SYM(t)->is_label= 1;
}
if(t->tag==CLABEL)
{SYM(t)->goto_count+= reference;
SYM(t)->scan_count--;
}
else if(t->tag==PRETURN)
SYM(t)->goto_count+= reference;
else
{if(t->tag==PINTEGER)
t->tag= CLABELN;
return;
}
DBG(dbgreturn,"Using label %s (%d) in line %d\n",SYM(t)->name,SYM(t)->goto_count,t->line_no);
}
	/*:158*/	/*183:*/
	#line 4547 "web2w.w"

static token*winsert_after(token*t,int tag,char*text)
{token*n;
DBG(dbgstmt,"Inserting token %s after %s in line %d\n",
tagname(tag),TAG(t),t->line_no);
n= new_token(tag);
n->next= t->next;
n->next->previous= n;
n->previous= t;
t->next= n;
n->sequence_no= t->sequence_no;
n->line_no= t->line_no;
n->text= text;
return n;
}
token*winsert_case(token*t,token*s)
{if(t==NULL)return NULL;
if(t->sequence_no==s->sequence_no&&t->pp_sp<s->pp_sp)
{DBG(dbgstmt,"Moving case in line %d (%d/%d!=%d)\n",
t->line_no,t->sequence_no,t->pp_sp,s->pp_sp);
while(t->previous->sequence_no==s->sequence_no&&
t->tag!=PMACRO)
t= t->previous;
}
return winsert_after(t->previous,CCASE,"case ");
}
	/*:183*/	/*190:*/
	#line 4747 "web2w.w"

void for_loop_variable(token*id,int line_no,int to,int direction)
{SYM(id)->is_int= 1;
DBG(dbgstmt,"\tLoop control variable %s: ""limit %d, direction %d in line %d\n",
SYM(id)->name,to,direction,line_no);
}
	/*:190*/	/*192:*/
	#line 4870 "web2w.w"

bool wneeds_semicolon(token*p)
{while(p!=NULL)
{switch(p->tag)
{case PCASE:case PBEGIN:case CIGNORE:return false;
case PSEMICOLON:
case CCASE:case PELSE:p= p->next;continue;
case PIF:case PWHILE:case PFOR:case PCOLON:p= p->previous;continue;
case PASSIGN:case PFUNCID:case PCALLID:case PREPEAT:
case PRETURN:case CRETURN:case CPROCRETURN:case PGOTO:case PEXIT:case CEMPTY:
default:return true;
}
}
return false;
}


void wsemicolon(token*p,token*t)
{t= wback(t);
if(t->tag!=PSEMICOLON&&t->tag!=CSEMICOLON&&t->tag!=PEND)
{if(wneeds_semicolon(p))
{DBG(dbgsemicolon,"inserting ; in line %d\n",t->line_no);
if(t->next->tag==ATSEMICOLON)
{t->next->tag= CSEMICOLON;t->next->text= ";";}
else
winsert_after(t,CSEMICOLON,";");
}
else if(t->next->tag!=ATSEMICOLON&&t->next->tag!=PSEMICOLON)
{
DBG(dbgsemicolon,"inserting @; in line %d\n",t->line_no);
winsert_after(t,ATSEMICOLON,"@;");
}

}
}
	/*:192*/	/*193:*/
	#line 4918 "web2w.w"

void wend(token*p,token*t)
{if(p->tag==PSEMICOLON&&p->next->tag==PCOLON&&
p->next->next->tag==CEMPTY&&p->next->previous->tag==CLABEL&&
p->next->previous->sym_no==exit_no)
{token*label= p->next->previous;
DBG(dbgreturn,"Trailing exit: found preceding line %d\n",t->line_no);
label->tag= CIGNORE;
SYM(label)->goto_count= -1000;
CHECK(label->next->tag==PCOLON,"Expected colon after label in line %d\n",label->line_no);
label->next->tag= CIGNORE;
p->next->tag= CIGNORE;
}
else
DBG(dbgreturn,"No trailing exit: found preceding line %d\n",t->line_no);
}

	/*:193*/	/*208:*/
	#line 5163 "web2w.w"

static bool wtail(token*t)
{CHECK(t!=NULL,"Unexpected NULL token while searching for tail statements");
switch(t->tag)
{case PSEMICOLON:case PELSE:case CCASE:
return wtail(t->next)&&wtail(t->previous);
case PCOLON:
return wtail(t->next);
case PRETURN:case CIGNORE:case CEMPTY:
return true;
case PASSIGN:case PCALLID:
case PFUNCID:case CRETURN:case CPROCRETURN:
case PWHILE:case PREPEAT:case PFOR:
case PEXIT:case PGOTO:
return false;
case PBEGIN:case PIF:case PCASE:
return wtail(t->previous);
default:ERROR("Unexpected tag %s while searching for tail statements",TAG(t));
}
}
	/*:208*/	/*210:*/
	#line 5203 "web2w.w"

void wreturn(token*t,int tail,token*link)
{CHECK(t!=NULL,"Unexpected NULL token while searching for return statements");
switch(t->tag)
{case PSEMICOLON:
if(t->next->tag==PRETURN)
wreturn(t->previous,true,t->next);
else
{wreturn(t->next,tail,link);
if(wtail(t->next))wreturn(t->previous,tail,link);
else wreturn(t->previous,false,NULL);
}
return;
case PCOLON:
wreturn(t->next,tail,link);
return;
case PASSIGN:case PCALLID:
case PRETURN:case PEXIT:case PGOTO:case CIGNORE:case CEMPTY:
return;
case PWHILE:case PREPEAT:case PFOR:
wreturn(t->previous,false,NULL);return;
case PELSE:case CCASE:
wreturn(t->next,tail,link);wreturn(t->previous,tail,link);return;
case PCASE:case PIF:case PBEGIN:
wreturn(t->previous,tail,link);return;
case PFUNCID:
if(tail)
{DBG(dbgreturn,"Converting assignment to function in line %d\n",t->line_no);
t->tag= CRETURN;IGN(t->next);
if(link!=NULL)
{link->sym_ptr->goto_count--;
t->sym_ptr= link->sym_ptr;
IGN(link),IGN(link->next);
DBG(dbgreturn,"Eliminating label %s (%d) in line %d\n",
link->sym_ptr->name,link->sym_ptr->goto_count,t->line_no);
}
}
return;
case CRETURN:
if(t->sym_ptr!=NULL){
t->sym_ptr->goto_count--;
DBG(dbgreturn,"Eliminating label %s (%d) in line %d\n",
t->sym_ptr->name,t->sym_ptr->goto_count,t->line_no);
}
return;
default:ERROR("Unexpected tag %s in line %d"
" while searching for return statements",TAG(t),t->line_no);
}
}
	/*:210*/	/*215:*/
	#line 5290 "web2w.w"

int wlocal_value(token*t,token*begin,token*end)
{int f_no= t->sym_no;
if(f_no!=x_over_n_no&&f_no!=xn_over_d_no)return 0;
DBG(dbgcweb,"Discovered function %s; in line %d\n",
SYM(t)->name,t->line_no);
CHGTAG(begin,PFBEGIN);begin->sym_no= f_no;
CHGTAG(end,PFEND);end->sym_no= f_no;
return 1;
}
	/*:215*/	/*218:*/
	#line 5338 "web2w.w"

void usage(void)
{fprintf(stderr,
"Usage: web2w [parameters] filename.web\n"
"Parameters:\n"
"\t -p   \t generate a Pascal output file\n"
"\t -o file   \t specify an output file name\n"
"\t -l   \t redirect stderr to a log file\n"
"\t -y   \t generate a trace while parsing Pascal\n"
"\t -u   \t generate macro names using upper case letters\n"
"\t -h   \t generate a header file section\n"
"\t -e file   \t file with extern identifiers\n"
	/*235:*/
	#line 5697 "web2w.w"

"\t -d XX\t hexadecimal debug value. OR together these values:\n"
"\t \t   \t XX=1   basic debugging\n"
"\t \t   \t XX=2   flex debugging\n"
"\t \t   \t XX=4   link debugging\n"
"\t \t   \t XX=8   token debugging\n"
"\t \t   \t XX=10  identifier debugging\n"
"\t \t   \t XX=20  Pascal tokens debugging\n"
"\t \t   \t XX=40  expansion debugging\n"
"\t \t   \t XX=80  bison debugging\n"
"\t \t   \t XX=100 Pascal parser debugging\n"
"\t \t   \t XX=200 cweb debugging\n"
"\t \t   \t XX=400 join debugging\n"
"\t \t   \t XX=800 string pool debugging\n"
"\t \t   \t XX=1000 Pascal statement debugging\n"
"\t \t   \t XX=2000 real division debugging\n"
"\t \t   \t XX=4000 macro debugging\n"
"\t \t   \t XX=8000 array debugging\n"
"\t \t   \t XX=10000 return debugging\n"
"\t \t   \t XX=20000 semicolon debugging\n"
"\t \t   \t XX=40000 break debugging\n"
"\t \t   \t XX=80000 type debugging\n"
	/*:235*/
	#line 5350 "web2w.w"

);
exit(1);
}
	/*:218*/
	#line 1090 "web2w.w"

int main(int argc,char*argv[])
{	/*220:*/
	#line 5364 "web2w.w"

{int mk_logfile= 0,mk_pascal= 0,baselength= 0;
char*w_file_name= NULL;
ww__flex_debug= 0;
pp_debug= 0;
if(argc<2)usage();
argv++;
while(*argv!=NULL)
{if((*argv)[0]=='-')
{char option= (*argv)[1];
switch(option)
{default:usage();
case'p':mk_pascal= 1;break;
case'o':argv++;if(*argv==NULL||*argv[0]=='-')usage();
w_file_name= *argv;break;
case'l':mk_logfile= 1;break;
case'y':pp_debug= 1;break;
case'u':uppercase= 1;break;
case'h':header_section= 1;break;
case'e':argv++;if(*argv==NULL||*argv[0]=='-')usage();
exports_name= *argv;break;
case'd':
argv++;if(*argv==NULL||*argv[0]=='-')usage();
debugflags= strtol(*argv,NULL,16);
if(debugflags&dbgflex)ww__flex_debug= 1;
if(debugflags&dbgbison)pp_debug= 1;
break;
}
}
else
{strncpy(basename,*argv,MAX_NAME-1);
baselength= strlen(basename);
if(baselength>4&&strncmp(basename+baselength-4,".web",4)==0)
baselength= baselength-4;
basename[baselength]= 0;
if(*(argv+1)!=NULL)usage();
}
argv++;
}
	/*225:*/
	#line 5448 "web2w.w"

if(mk_logfile)
{
strcat(basename,".log");
logfile= freopen(basename,"w",stderr);
if(logfile==NULL)
{fprintf(stderr,"Unable to open logfile %s",basename);
logfile= stderr;
}
ww_out= logfile;
basename[baselength]= 0;
}
else
{logfile= stderr;
ww_out= stderr;
}
strcat(basename,".web");
ww_in= fopen(basename,"r");
if(ww_in==NULL)ERROR("Unable to open input file %s",basename);
basename[baselength]= 0;
if(w_file_name==NULL)
{strcat(basename,".w");
w_file_name= basename;
}
w= fopen(w_file_name,"w");
if(w==NULL)ERROR("Unable to open output file %s",w_file_name);
basename[baselength]= 0;
if(mk_pascal)
{strcat(basename,".pas");
pp_out= fopen(basename,"w");
if(pp_out==NULL)ERROR("Unable to open pp_out file %s",basename);
basename[baselength]= 0;
}
	/*:225*/
	#line 5403 "web2w.w"

}
	/*:220*/
	#line 1092 "web2w.w"

	/*5:*/
	#line 1128 "web2w.w"

	/*23:*/
	#line 1349 "web2w.w"

first_token= last_token= new_token(HEAD);
first_token->text= "";
	/*:23*/	/*61:*/
	#line 1946 "web2w.w"

program= first_token;
	/*:61*/	/*103:*/
	#line 2865 "web2w.w"

predefine("put",PPROCID,0);
predefine("get",PPROCID,0);
predefine("reset",PPROCID,0);
predefine("rewrite",PPROCID,0);
predefine("abs",PFUNCID,0);
predefine("odd",PFUNCID,0);
predefine("eof",PFUNCID,0);
predefine("eoln",PFUNCID,0);
predefine("round",PFUNCID,0);
predefine("ord",PFUNCID,0);
predefine("chr",PFUNCID,0);
predefine("close",PPROCID,0);
predefine("read",PPROCID,0);
predefine("read_ln",PPROCID,0);
predefine("write",PPROCID,0);
predefine("write_ln",PPROCID,0);
predefine("break",PPROCID,0);
predefine("break_in",PPROCID,0);
predefine("erstat",PFUNCID,0);
predefine("false",PCONSTID,0);
predefine("true",PCONSTID,1);
	/*:103*/	/*128:*/
	#line 3338 "web2w.w"

float_constant_no= predefine("float_constant",ID,0);
	/*:128*/	/*143:*/
	#line 3630 "web2w.w"

printn_no= predefine("printn",PPROCID,0);
print_no= predefine("print",PPROCID,0);
print_esc_no= predefine("print_esc",PPROCID,0);
printn_esc_no= predefine("printn_esc",PPROCID,0);
print_nl_no= predefine("print_nl",PPROCID,0);
scan_keyword_no= predefine("scan_keyword",PFUNCID,0);
	/*:143*/	/*161:*/
	#line 4053 "web2w.w"

exit_no= sym_no("exit");
	/*:161*/	/*213:*/
	#line 5275 "web2w.w"

x_over_n_no= sym_no("x_over_n");
xn_over_d_no= sym_no("xn_over_d");
	/*:213*/
	#line 1129 "web2w.w"

ww_lex();
	/*70:*/
	#line 2170 "web2w.w"

TOK("",ATP);PROGRAM;PUSH;TOK("",PEOF);TOK("",WEBEOF);POP;
SYM_PTR("main")->is_extern= 1;
SYM_PTR("\"TeXinputs:\"")->name= "\"TeXinputs/\"";
SYM_PTR("\"TeXfonts:\"")->name= "\"TeXfonts/\"";
	/*:70*/	/*71:*/
	#line 2180 "web2w.w"

if(debugflags&dbgtoken)
{token*t= first_token;
while(t!=NULL){MESSAGE(THE_TOKEN(t));t= t->next;}
}
if(debugflags&dbgid)
{int i;
for(i= free_symbols;i<MAX_SYMBOLS;i++)
MESSAGE("symbol[%d]=%s (%s)\n",i,symbols[i].name,tagname(symbols[i].tag));
}
	/*:71*/	/*86:*/
	#line 2463 "web2w.w"

SYM_PTR("below_display_skip")->use_count= 1;
SYM_PTR("below_display_short_skip")->use_count= 1;
SYM_PTR("top_skip")->use_count= 1;
SYM_PTR("tab_skip")->use_count= 1;
SYM_PTR("thin_mu_skip")->use_count= 1;
SYM_PTR("med_mu_skip")->use_count= 1;
SYM_PTR("thick_mu_skip")->use_count= 1;
SYM_PTR("toks")->use_count= 1;
SYM_PTR("output_penalty")->use_count= 1;
SYM_PTR("TeX_banner")->use_count= 1;
	/*:86*/	/*229:*/
	#line 5523 "web2w.w"

if(exports_name!=NULL)
{int c,n= 0;
char s[MAX_NAME];
int k;
FILE*e= fopen(exports_name,"r");
if(e==NULL)ERROR("Unable to open extern name file %s",exports_name);
SYM_PTR("main")->is_extern= 0;
c= getc(e);
while(c!=EOF)
{n++;
k= 0;
while(c!=EOF)
{if(c=='\n'){c= getc(e);break;}
else if(c=='\r')c= getc(e);
else if(!isalnum(c)&&c!='_')
ERROR("Unexpected character %c(%x) in %s, line %d\n",
c,c,exports_name,n);
else
{s[k++]= c;
if(k>=MAX_NAME)
ERROR("extern name too long in %s, line %d\n",
exports_name,n);
c= getc(e);
}
}
if(k>0)
{int i;
s[k]= 0;
i= sym_no(s);
symbol_table[i]->is_extern= extern_symbols;
extern_symbols= i;
DBG(dbgid,"Extern symbol %s\n",symbol_table[i]->name);
}
}
fclose(e);
}
	/*:229*/
	#line 1131 "web2w.w"

	/*66:*/
	#line 2033 "web2w.w"

{symbol*s;
for(s= nmacros;s!=NULL;s= s->link)
{token*t,*start,*end;
t= s->eq->next;
while(!is_pascal(t))t= t->next;
start= t;
if(t->tag==PINTEGER||t->tag==OCTAL||t->tag==HEX||
t->tag==PCHAR||t->tag==PSTRING||t->tag==CHAR||t->tag==STRING||
t->tag==NMACRO)t= t->next;
end= NULL;
while(t!=s->eq->link)
{if(t->tag==MLEFT)t= t->link;
else if(is_pascal(t))end= t;
t= t->next;
}
if(end!=NULL)
{winsert_after(start->previous,POPEN,"(");
winsert_after(end,PCLOSE,")");
DBG(dbgmacro,"Adding parentheses for numeric macro %s in line %d\n",s->name,s->eq->line_no);
}
}
}
	/*:66*/
	#line 1132 "web2w.w"

	/*67:*/
	#line 2066 "web2w.w"

{symbol*s;
for(s= omacros;s!=NULL;s= s->link)
{bool convertible= false;
token*t;
t= s->eq->next;
if(t->tag==PPLUS||t->tag==PMINUS)t= t->next;
if(t->tag==PINTEGER||t->tag==OCTAL||t->tag==HEX||
t->tag==NMACRO||(t->tag==OMACRO&&SYM(t)->tag==NMACRO))
{convertible= true;
for(t= t->next;convertible&&t!=s->eq->link;t= t->next)
if(!is_pascal(t))continue;
else if(t->tag==MLEFT)t= t->link;
else convertible= false;
}
if(convertible)
{t= s->eq->previous;
t->tag= NMACRO;
s->tag= NMACRO;
DBG(dbgmacro,"Converting %s from ordinary to numeric macro in line %d\n",s->name,t->line_no);
}
}
}
	/*:67*/
	#line 1133 "web2w.w"

	/*:5*/
	#line 1093 "web2w.w"

	/*99:*/
	#line 2800 "web2w.w"

program= first_token->link;
pp_push(program->link,program->next,0,NULL);
pp_parse();
	/*69:*/
	#line 2142 "web2w.w"

{symbol*s;
for(s= pmacros;s!=NULL;s= s->link)
if(s->arity<s->arg_count)
{token*t= tail_call(s->eq);
s->type= t;
if(t->tag==PMACRO)
{DBG(dbgmacro,"Tail call of %s (%d,%d) found in macro %s (%d) in line %d\n",
SYM(t)->name,SYM(t)->use_count,SYM(t)->scan_count,s->name,
s->use_count,t->line_no);
SYM(t)->use_count-= s->use_count;
SYM(t)->scan_count--;
}
}
}
	/*:69*/
	#line 2804 "web2w.w"

	/*:99*/
	#line 1094 "web2w.w"

	/*114:*/
	#line 3085 "web2w.w"

	/*133:*/
	#line 3425 "web2w.w"

SYM_PTR("switch")->name= "get_cur_chr";
SYM_PTR("continue")->name= "resume";
SYM_PTR("int")->name= "i";
SYM_PTR("register")->name= "internal_register";

SYM_PTR("exit")->name= "end";
SYM_PTR("free")->name= "is_free";
SYM_PTR("write")->name= "pascal_write";
SYM_PTR("read")->name= "pascal_read";
SYM_PTR("close")->name= "pascal_close";

SYM_PTR("xclause")->name= "else";
SYM_PTR("remainder")->name= "rem";
	/*:133*/
	#line 3086 "web2w.w"

wprint_to(first_token,last_token);
	/*227:*/
	#line 5494 "web2w.w"

if(header_section)
{wputs("\n@ Generating the header file.\n");
wputs("@(");wputs(basename);wputs(".h@>=\n"
"@h\n"
"enum {@+@<Constants in the outer block@>@+};\n"
"@<Types in the outer block@>@;\n");
	/*230:*/
	#line 5576 "web2w.w"

while(extern_symbols>0)
{symbol*x= symbol_table[extern_symbols];
token*t= x->type;
if(t==NULL)
ERROR("Symbol %s in file %s was never defined",x->name,exports_name);
else
{wprint("extern");
if(t->tag==CMAIN)
wprint("int main(void)");
if(t->tag==PPROCEDURE)
	/*198:*/
	#line 4975 "web2w.w"

{wprint("void");
t= wprint_to(t->next,t->link);
if(t->tag!=POPEN)wputs("(void)");
else t= wprint_to(t,t->link->next);
}
	/*:198*/
	#line 5587 "web2w.w"

else if(t->tag==PFUNCTION)
	/*207:*/
	#line 5132 "web2w.w"

{token*param= t->link;
token*type;
if(param->tag==POPEN)type= param->link->link;
else type= param;
wprint_to(type,type->link);
wprint_to(t->next,t->link);
if(param->tag!=POPEN)wputs("(void)");
else wprint_to(param,param->link->next);
t= type->link;
}
	/*:207*/
	#line 5589 "web2w.w"

else if(t->tag==PDEFVARID||t->tag==PID)
{varlist= t;
token*array_type= NULL;
	/*168:*/
	#line 4192 "web2w.w"

symbol*s;
int is_extern= 0,string_type= 0,is_int= 0,is_global= 0;
do{s= t->sym_ptr;
is_global|= s->is_global;
is_extern|= s->is_extern;
is_int|= s->is_int;
if(s->is_string>0)
{DBG(dbgstring,"Variable %s string (%d) defined in line %d\n",
s->name,s->is_string,t->line_no);
string_type= 1;
}
t= t->link;
}while(t->tag==PID);
	/*:168*/
	#line 5593 "web2w.w"

if(!is_extern)ERROR("Static variables become non static in line %d\n",t->line_no);
if(t->next->tag==PARRAY)
array_type= t->next;
else if(t->next->tag==PPACKED&&t->next->next->tag==PARRAY)
array_type= t->next->next;
if(array_type!=NULL)
{token*element_type= array_type->link->link->link->link;
wprint_to(element_type,element_type->link);
if(x->is_zero_based)
{wprint(x->name);wputs("[]");}
else
{wputs(" *const ");wprint(x->name);}
}
else
{	/*169:*/
	#line 4213 "web2w.w"

if(string_type)wprint("char *");
else if(is_int)wprint("int");
else wprint_to(t,t->link);
	/*:169*/
	#line 5608 "web2w.w"

wprint(x->name);
}
}
wputs(";\n");
}
extern_symbols= x->is_extern;
}
	/*:230*/
	#line 5501 "web2w.w"

}
	/*:227*/
	#line 3088 "web2w.w"

	/*:114*/
	#line 1095 "web2w.w"

	/*13:*/
	#line 1282 "web2w.w"

DBG(dbgbasic,"free tokens = %d\n",free_tokens);
	/*:13*/	/*32:*/
	#line 1467 "web2w.w"

DBG(dbgbasic,"free strings = %d\n",free_strings);
	/*:32*/	/*37:*/
	#line 1552 "web2w.w"

DBG(dbgbasic,"free symbols = %d\n",free_symbols);
	/*:37*/	/*55:*/
	#line 1867 "web2w.w"

DBG(dbgbasic,"free modules = %d\n",free_modules);
	/*:55*/
	#line 1096 "web2w.w"

return 0;
}
	/*:2*/
